use std::marker::PhantomData;

use std::sync::{Arc, RwLock};

use crate::assets::assetmanager::AssetManager;

use crate::{assets::entity_object_manager::EntityObjectManager, prelude::*};

use anyhow::Result;

pub trait AssetLoader {
    fn load_entity(&mut self, assets: AssetHandler<'_>, entity_file: &str) -> Result<EntityObject>;
}

#[derive(Clone, Copy)]
pub struct AssetHandler<'a> {
    pub(crate) asset_manager: &'a RwLock<AssetManager>,

    pub(crate) context: &'a Arc<Context>,

    pub(crate) resource_base_path: &'a String,

    pub(crate) entity_object_manager: &'a EntityObjectManager,

    pub(crate) phantom_data: PhantomData<&'a ()>,
}

impl<'a> AssetHandler<'a> {
    pub fn context(&self) -> &Arc<Context> {
        self.context
    }

    pub fn update_file_cache(&self) -> Result<()> {
        self.asset_manager.write().unwrap().update_files()
    }

    pub fn preload_assets<'b>(&self, files: impl IntoIterator<Item = &'b &'b str>) -> Result<()> {
        self.asset_manager
            .read()
            .unwrap()
            .async_load_gltf_file(self.context, files)
    }

    pub fn create_cache<'b>(
        &self,
        names: impl IntoIterator<Item = &'b &'b str> + Clone,
    ) -> Result<AssetCache> {
        self.asset_manager
            .write()
            .unwrap()
            .create_cache_with(self.context, names)
    }

    pub fn insert_cache(&self, cache: &AssetCache) -> Result<()> {
        self.asset_manager
            .read()
            .unwrap()
            .insert_cached_assets(cache)
    }

    pub fn clear(&self) -> Result<()> {
        self.asset_manager.write().unwrap().clear()
    }

    pub fn gltf_files(&self) -> Result<Vec<String>> {
        Ok(self.asset_manager.read().unwrap().gltf_files().clone())
    }

    pub fn create_entity(&self, asset_file: &str) -> Result<EntityObject> {
        self.asset_manager.write().unwrap().load_asset_file(
            self.context,
            self.entity_object_manager,
            asset_file,
        )
    }

    pub fn create_entity_absolute(&self, asset_path: &str) -> Result<EntityObject> {
        self.asset_manager
            .write()
            .unwrap()
            .load_asset_file_absolute(self.context, self.entity_object_manager, asset_path)
    }

    pub fn empty_entity(&self) -> EntityObject {
        self.entity_object_manager.create_entity()
    }

    pub fn resource_base_path(&self) -> &str {
        self.resource_base_path
    }

    pub(crate) fn get_entity_id(&self) -> u32 {
        self.entity_object_manager.fetch_add_entity_id()
    }
}
