use crate::prelude::*;

use std::sync::{
    atomic::{AtomicBool, Ordering::SeqCst},
    Arc, Mutex,
};

use anyhow::Result;

pub struct ValidCommandBuffers {
    combined: Vec<(Arc<CommandBuffer>, AtomicBool)>,
}

impl ValidCommandBuffers {
    pub fn new(
        device: &Arc<Device>,
        queue: &Arc<Mutex<Queue>>,
        image_count: usize,
    ) -> Result<Arc<ValidCommandBuffers>> {
        let mut valid_buffers = Vec::new();

        for _ in 0..image_count {
            let command_buffer =
                CommandBuffer::new_secondary().build(device.clone(), queue.clone())?;

            valid_buffers.push((command_buffer, AtomicBool::new(false)));
        }

        Ok(Arc::new(ValidCommandBuffers {
            combined: valid_buffers,
        }))
    }

    pub fn reset_all(&self) {
        for (_, valid) in &self.combined {
            valid.store(false, SeqCst);
        }
    }

    pub fn is_valid(&self, index: usize) -> bool {
        let (_, valid) = &self.combined[index];
        valid.load(SeqCst)
    }

    pub fn current(&self, index: usize) -> (&Arc<CommandBuffer>, &AtomicBool) {
        let (cb, valid) = &self.combined[index];
        (cb, valid)
    }
}
