use crate::prelude::*;

use anyhow::Result;
use std::sync::Arc;

pub struct RasterizerPipelines;

impl RasterizerPipelines {
    #[allow(clippy::too_many_arguments)]
    pub fn create_default_pipeline<T: VertexInputDescription>(
        device: Arc<Device>,
        pipeline_layout: &Arc<PipelineLayout>,
        vertex_shader: Arc<ShaderModule<shader_type::Vertex>>,
        fragment_shader: Arc<ShaderModule<shader_type::Fragment>>,
        renderpass: &Arc<RenderPass>,
        attachment_info: &[bool],
        sub_pass_index: u32,
    ) -> Result<Arc<Pipeline>> {
        let blend_attachments: Vec<VkPipelineColorBlendAttachmentState> = attachment_info
            .iter()
            .map(|info| Self::color_blend_attachment(*info))
            .collect();

        let pipeline = Pipeline::new_graphics()
            .set_vertex_shader::<T>(vertex_shader)
            .set_fragment_shader(fragment_shader)
            .input_assembly(VK_PRIMITIVE_TOPOLOGY_TRIANGLE_LIST, false)
            .default_depth_stencil(true, false)
            .default_color_blend(blend_attachments)
            .default_rasterization(VK_CULL_MODE_NONE, VK_FRONT_FACE_COUNTER_CLOCKWISE)
            .default_multisample(VK_SAMPLE_COUNT_1_BIT)
            .enable_rasterization_order(VK_RASTERIZATION_ORDER_RELAXED_AMD)
            .build(device, pipeline_layout, renderpass, sub_pass_index)?;

        Ok(pipeline)
    }

    #[allow(clippy::too_many_arguments)]
    pub fn multi_sample_pipeline<T: VertexInputDescription>(
        device: Arc<Device>,
        pipeline_layout: &Arc<PipelineLayout>,
        vertex_shader: Arc<ShaderModule<shader_type::Vertex>>,
        fragment_shader: Arc<ShaderModule<shader_type::Fragment>>,
        renderpass: &Arc<RenderPass>,
        attachment_info: &[bool],
        sub_pass_index: u32,
        sample_count: VkSampleCountFlags,
        cull_mode: VkCullModeFlags,
    ) -> Result<Arc<Pipeline>> {
        let blend_attachments: Vec<VkPipelineColorBlendAttachmentState> = attachment_info
            .iter()
            .map(|info| Self::color_blend_attachment(*info))
            .collect();

        let pipeline = Pipeline::new_graphics()
            .set_vertex_shader::<T>(vertex_shader)
            .set_fragment_shader(fragment_shader)
            .input_assembly(VK_PRIMITIVE_TOPOLOGY_TRIANGLE_LIST, false)
            .default_depth_stencil(true, false)
            .default_color_blend(blend_attachments)
            .default_rasterization(cull_mode, VK_FRONT_FACE_COUNTER_CLOCKWISE)
            .default_multisample(sample_count)
            .enable_rasterization_order(VK_RASTERIZATION_ORDER_RELAXED_AMD)
            .build(device, pipeline_layout, renderpass, sub_pass_index)?;

        Ok(pipeline)
    }

    #[allow(clippy::too_many_arguments)]
    pub fn create_shadow_pipeline<T: VertexInputDescription>(
        device: Arc<Device>,
        pipeline_layout: &Arc<PipelineLayout>,
        vertex_shader: Arc<ShaderModule<shader_type::Vertex>>,
        fragment_shader: Arc<ShaderModule<shader_type::Fragment>>,
        renderpass: &Arc<RenderPass>,
        attachment_info: &[bool],
        sub_pass_index: u32,
    ) -> Result<Arc<Pipeline>> {
        let blend_attachments: Vec<VkPipelineColorBlendAttachmentState> = attachment_info
            .iter()
            .map(|info| Self::color_blend_attachment(*info))
            .collect();

        // init depth/stencil state
        let stencil_op_state = VkStencilOpState {
            failOp: VK_STENCIL_OP_KEEP,
            passOp: VK_STENCIL_OP_KEEP,
            depthFailOp: VK_STENCIL_OP_KEEP,
            compareOp: VK_COMPARE_OP_ALWAYS,
            compareMask: 0,
            writeMask: 0,
            reference: 0,
        };

        let pipeline = Pipeline::new_graphics()
            .set_vertex_shader::<T>(vertex_shader)
            .set_fragment_shader(fragment_shader)
            .input_assembly(VK_PRIMITIVE_TOPOLOGY_TRIANGLE_LIST, false)
            .custom_depth_stencil(
                true,
                true,
                VK_COMPARE_OP_LESS_OR_EQUAL,
                false,
                false,
                stencil_op_state.clone(),
                stencil_op_state,
                0.0,
                0.0,
            )
            .default_color_blend(blend_attachments)
            .custom_rasterization(
                false,
                false,
                VK_POLYGON_MODE_FILL,
                VK_CULL_MODE_BACK_BIT,
                VK_FRONT_FACE_COUNTER_CLOCKWISE,
                true,
                1.25,
                0.0,
                1.75,
                1.0,
            )
            .default_multisample(VK_SAMPLE_COUNT_1_BIT)
            .build(device, pipeline_layout, renderpass, sub_pass_index)?;

        Ok(pipeline)
    }
}

// private
impl RasterizerPipelines {
    fn color_blend_attachment(enabled: bool) -> VkPipelineColorBlendAttachmentState {
        if enabled {
            VkPipelineColorBlendAttachmentState::default()
        } else {
            VkPipelineColorBlendAttachmentState {
                blendEnable: VK_FALSE,
                srcColorBlendFactor: VK_BLEND_FACTOR_ZERO,
                dstColorBlendFactor: VK_BLEND_FACTOR_ZERO,
                colorBlendOp: VK_BLEND_OP_ADD,
                srcAlphaBlendFactor: VK_BLEND_FACTOR_ZERO,
                dstAlphaBlendFactor: VK_BLEND_FACTOR_ZERO,
                alphaBlendOp: VK_BLEND_OP_ADD,
                colorWriteMask: VK_COLOR_COMPONENT_R_BIT
                    | VK_COLOR_COMPONENT_G_BIT
                    | VK_COLOR_COMPONENT_B_BIT
                    | VK_COLOR_COMPONENT_A_BIT,
            }
        }
    }
}
