use super::super::shared::shadercompiler::{
    CompileOptions, ResolvedInclude, ShaderCompiler, ShaderTypeConverter, Shaders, SpirvVersion,
};

use crate::prelude::*;

use anyhow::Result;
use shaderc::IncludeType;
use std::sync::Arc;

const SHARED_SHADER_HEADER: &str = include_str!("../shaders/raytracing/hardware/global_data.glsl");
const SHARED_PBR_HEADER: &str = include_str!("../shaders/pbr.glsl");

fn shader_includes(
    name: &str,
    _include_type: IncludeType,
    _: &str,
    _: usize,
) -> std::result::Result<ResolvedInclude, String> {
    if name.contains("global_data.glsl") {
        Ok(ResolvedInclude {
            resolved_name: name.to_string(),
            content: SHARED_SHADER_HEADER.to_string(),
        })
    } else if name.contains("pbr.glsl") {
        Ok(ResolvedInclude {
            resolved_name: name.to_string(),
            content: SHARED_PBR_HEADER.to_string(),
        })
    } else {
        unreachable!()
    }
}

fn set_spv_version(compiler_options: &mut CompileOptions<'_>) {
    compiler_options.set_target_spirv(SpirvVersion::V1_4);
}

pub struct DefaultHardwareRayTracingShader {
    pub ray_gen: Arc<ShaderModule<shader_type::RayGeneration>>,
    pub miss: Arc<ShaderModule<shader_type::Miss>>,
    pub closest_hit: Arc<ShaderModule<shader_type::ClosestHit>>,
    pub shadow_hit: Arc<ShaderModule<shader_type::AnyHit>>,
    pub shadow_miss: Arc<ShaderModule<shader_type::Miss>>,
}

impl DefaultHardwareRayTracingShader {
    pub fn new(device: &Arc<Device>, recursion_depth: u32) -> Result<Self> {
        Ok(Self {
            ray_gen: Self::simple_ray_gen(device)?,
            miss: Self::simple_miss(device)?,
            closest_hit: Self::simple_closest_hit(device, recursion_depth)?,
            shadow_hit: Self::shadow_hit(device)?,
            shadow_miss: Self::shadow_miss(device)?,
        })
    }

    #[inline]
    fn simple_ray_gen(
        device: &Arc<Device>,
    ) -> Result<Arc<ShaderModule<shader_type::RayGeneration>>> {
        let rgen_text = include_str!("../shaders/raytracing/hardware/default/raygen.rgen");

        let compiler = ShaderCompiler::new(vec![(rgen_text, "raygen.rgen")])?;

        compiler.compile::<_, 0>(device, &[], set_spv_version)
    }

    #[inline]
    fn simple_miss(device: &Arc<Device>) -> Result<Arc<ShaderModule<shader_type::Miss>>> {
        let miss_text = include_str!("../shaders/raytracing/hardware/default/miss.rmiss");

        let compiler = ShaderCompiler::new(vec![(miss_text, "miss.rmiss")])?;

        compiler.compile::<_, 0>(device, &[], set_spv_version)
    }

    #[inline]
    fn shadow_miss(device: &Arc<Device>) -> Result<Arc<ShaderModule<shader_type::Miss>>> {
        let shadow_miss_text =
            include_str!("../shaders/raytracing/hardware/default/shadowmiss.rmiss");

        let compiler = ShaderCompiler::new(vec![(shadow_miss_text, "shadow_miss.rmiss")])?;

        compiler.compile::<_, 0>(device, &[], set_spv_version)
    }

    #[inline]
    fn simple_closest_hit(
        device: &Arc<Device>,
        recursion_depth: u32,
    ) -> Result<Arc<ShaderModule<shader_type::ClosestHit>>> {
        let closest_hit_text =
            include_str!("../shaders/raytracing/hardware/default/closesthit.rchit");

        let compiler = ShaderCompiler::new(vec![(closest_hit_text, "closesthit.rchit")])?;

        Ok(compiler.compile_with_includes(
            device,
            &[("RECURSION_DEPTH", Some(&recursion_depth.to_string()))],
            shader_includes,
            set_spv_version,
        )?[0]
            .clone())
    }

    #[inline]
    fn shadow_hit(device: &Arc<Device>) -> Result<Arc<ShaderModule<shader_type::AnyHit>>> {
        let any_hit_text = include_str!("../shaders/raytracing/hardware/default/shadow_hit.rahit");

        let compiler = ShaderCompiler::new(vec![(any_hit_text, "shadow_hit.rahit")])?;

        Ok(
            compiler.compile_with_includes(device, &[], shader_includes, set_spv_version)?[0]
                .clone(),
        )
    }
}

pub struct RTInAWeekendShader {
    pub ray_gen: Arc<ShaderModule<shader_type::RayGeneration>>,
    pub miss: Arc<ShaderModule<shader_type::Miss>>,
    pub closest_hit: Arc<ShaderModule<shader_type::ClosestHit>>,
}

impl RTInAWeekendShader {
    pub fn new(device: &Arc<Device>, recursion_depth: u32) -> Result<Self> {
        Ok(Self {
            ray_gen: Self::ray_gen(device)?,
            miss: Self::miss(device)?,
            closest_hit: Self::closest_hit(device, recursion_depth)?,
        })
    }

    #[inline]
    fn ray_gen(device: &Arc<Device>) -> Result<Arc<ShaderModule<shader_type::RayGeneration>>> {
        let shader_text =
            include_str!("../shaders/raytracing/hardware/rt_in_a_weekend/raygen.rgen");

        let compiler = ShaderCompiler::new(vec![(shader_text, "raygen.rgen")])?;

        Ok(
            compiler.compile_with_includes(device, &[], shader_includes, set_spv_version)?[0]
                .clone(),
        )
    }

    #[inline]
    fn miss(device: &Arc<Device>) -> Result<Arc<ShaderModule<shader_type::Miss>>> {
        let shader_text = include_str!("../shaders/raytracing/hardware/rt_in_a_weekend/miss.rmiss");

        let compiler = ShaderCompiler::new(vec![(shader_text, "miss.rmiss")])?;

        compiler.compile::<_, 0>(device, &[], set_spv_version)
    }

    #[inline]
    fn closest_hit(
        device: &Arc<Device>,
        recursion_depth: u32,
    ) -> Result<Arc<ShaderModule<shader_type::ClosestHit>>> {
        let shader_text =
            include_str!("../shaders/raytracing/hardware/rt_in_a_weekend/closest_hit.rchit");

        let compiler = ShaderCompiler::new(vec![(shader_text, "closest_hit.rchit")])?;

        Ok(compiler.compile_with_includes(
            device,
            &[("RECURSION_DEPTH", Some(&recursion_depth.to_string()))],
            shader_includes,
            set_spv_version,
        )?[0]
            .clone())
    }
}

pub struct RayTracingShader {}

impl RayTracingShader {
    pub fn animation(device: &Arc<Device>) -> Result<Shaders> {
        let cs_text = include_str!("../shaders/raytracing/animation.comp");

        let compiler = ShaderCompiler::new(vec![(cs_text, "shader/animation.comp")])?;

        compiler.compile_stages(
            true,
            device,
            &[("BONE_COUNT", Some(asset::BONE_COUNT.to_string()))],
            [
                &[],
                &["COLOR_UVS"],
                &["NORMALS"],
                &["NORMALS", "BONES"],
                &["NORMALS", "COLOR"],
                &["NORMALS", "COLOR_UVS"],
                &["NORMALS", "COLOR_UVS", "BONES"],
                &["NORMALS", "COLOR_UVS", "NORMAL_UVS"],
                &["NORMALS", "COLOR_UVS", "NORMAL_UVS", "BONES"],
            ],
        )
    }

    pub fn acceleration_structure(
        device: &Arc<Device>,
        triangles_per_as: u32,
    ) -> Result<Arc<ShaderModule<shader_type::Compute>>> {
        let cs_text = include_str!("../shaders/raytracing/software/acceleration_structure.comp");

        let compiler = ShaderCompiler::new(vec![(cs_text, "shader/acceleration_structure.comp")])?;

        compiler.compile::<_, 0>(
            device,
            &[("TRIANGLES_PER_AS", Some(triangles_per_as.to_string()))],
            set_spv_version,
        )
    }

    pub fn acceleration_tree(
        device: &Arc<Device>,
    ) -> Result<Arc<ShaderModule<shader_type::Compute>>> {
        let cs_text = include_str!("../shaders/raytracing/software/acceleration_tree.comp");

        let compiler = ShaderCompiler::new(vec![(cs_text, "shader/acceleration_tree.comp")])?;

        compiler.compile::<_, 0>(device, &[], set_spv_version)
    }

    pub fn raytracing(
        device: &Arc<Device>,
        max_samplers: u32,
        stack_size: u32,
    ) -> Result<Arc<ShaderModule<shader_type::Compute>>> {
        let cs_text = include_str!("../shaders/raytracing/software/software_raytracing.comp");

        let compiler = ShaderCompiler::new(vec![(cs_text, "shader/software_raytracing.comp")])?;

        compiler.compile::<_, 0>(
            device,
            &[
                ("MAX_SAMPLERS", Some(max_samplers.to_string())),
                ("STACK_SIZE", Some(stack_size.to_string())),
            ],
            set_spv_version,
        )
    }
}
