use cgmath;
use cgmath::Vector2;
use engine::prelude::*;
use entity_manager::*;

use super::{MOVEMENT_MAX_THRESHOLD, MOVEMENT_MIN_THRESHOLD};

use std::time::Duration;

pub struct Movement {
    direction: Vector2<f32>,
    pub movement_speed: f32,

    lock: bool,

    pub(crate) base_movement_speed: f32,

    pub current_distance: Option<f32>,

    audio: Option<UnsafeComponentStore<Audio>>,
}

impl Clone for Movement {
    fn clone(&self) -> Self {
        Movement {
            direction: self.direction,
            movement_speed: self.base_movement_speed,
            base_movement_speed: self.base_movement_speed,

            lock: false,

            current_distance: None,

            audio: None,
        }
    }
}

impl Movement {
    pub(crate) fn new(speed: f32, audio: Option<&mut Audio>) -> Self {
        Self {
            direction: Vector2::zero(),
            movement_speed: speed,
            base_movement_speed: speed,

            lock: false,

            current_distance: None,

            audio: audio.map(|audio| audio.into()),
        }
    }

    pub fn post_setup(&mut self, audio: Option<&mut Audio>) {
        self.audio = audio.map(|audio| audio.into());
    }

    pub fn set_direction(&mut self, direction: Vector2<f32>) {
        self.direction = self.check_direction(direction);
    }

    pub fn direction(&self) -> Vector2<f32> {
        self.direction
    }

    pub fn movement_speed(&self) -> f32 {
        self.movement_speed
    }

    pub fn base_movement_speed(&self) -> f32 {
        self.base_movement_speed
    }

    pub fn is_locked(&self) -> bool {
        self.lock
    }

    fn check_direction(&mut self, mut direction: Vector2<f32>) -> Vector2<f32> {
        let length = direction.magnitude();

        if length < MOVEMENT_MIN_THRESHOLD {
            if let Some(ptr) = &mut self.audio {
                let audio = unsafe { ptr.as_mut() };

                audio.stop_looping().unwrap();
            }

            direction = Vector2::zero();
        } else {
            if let Some(ptr) = &mut self.audio {
                let audio = unsafe { ptr.as_mut() };

                audio.play(SOUND_MOVE_KEY, true).unwrap();
            }

            if length > MOVEMENT_MAX_THRESHOLD {
                direction /= length;
            }
        }

        direction
    }

    pub fn has_direction(&self) -> bool {
        self.direction != Vector2::zero()
    }

    pub fn process_move(&mut self, time_delta: Duration) -> Vector2<f32> {
        self.distance(time_delta) * self.direction
    }

    pub fn distance(&mut self, time_delta: Duration) -> f32 {
        self.movement_speed * time_delta.as_secs_f32()
    }

    pub fn setup_after_clone(&mut self, multi_mut: &mut MultiMut<'_>) {
        let audio = multi_mut.get::<Audio>();

        self.audio = audio.map(UnsafeComponentStore::from).ok();
    }
}

impl EntityComponent for Movement {
    fn name(&self) -> &str {
        Self::debug_name()
    }
}

impl ComponentDebug for Movement {
    fn debug_name() -> &'static str {
        "Movement"
    }
}
