use std::{fs, mem::ManuallyDrop};

use assetpath::AssetPath;
use mlua::{FromLua, FromLuaMulti, Function, IntoLua, IntoLuaMulti, Lua, Table as MLuaTable};

pub struct Globals<'a> {
    table: MLuaTable<'a>,
    lua: &'a Lua,
}

impl<'a> Globals<'a> {
    pub fn set<K: IntoLua<'a>, V: IntoLua<'a>>(
        &self,
        key: K,
        value: V,
    ) -> Result<(), anyhow::Error> {
        self.table.set(key, value)?;

        Ok(())
    }

    pub fn get<K: IntoLua<'a>, V: FromLua<'a>>(&self, key: K) -> Result<V, anyhow::Error> {
        Ok(self.table.get(key)?)
    }

    pub fn create_table(&self) -> Result<Table<'_>, anyhow::Error> {
        Ok(Table {
            table: self.lua.create_table()?,
        })
    }

    pub fn set_table(&self, name: &str, table: Table<'_>) -> Result<(), anyhow::Error> {
        self.set(name, table.table)?;

        Ok(())
    }

    pub fn get_table(&self, name: &str) -> Result<Table<'_>, anyhow::Error> {
        Ok(Table {
            table: self.get(name)?,
        })
    }
}

pub struct Table<'a> {
    table: MLuaTable<'a>,
}

impl<'a> Table<'a> {
    pub fn set<K: IntoLua<'a>, V: IntoLua<'a>>(
        &self,
        key: K,
        value: V,
    ) -> Result<(), anyhow::Error> {
        self.table.set(key, value)?;

        Ok(())
    }

    pub fn get<K: IntoLua<'a>, V: FromLua<'a>>(&self, key: K) -> Result<V, anyhow::Error> {
        Ok(self.table.get(key)?)
    }
}

pub struct LuaFunction {
    lua: Option<&'static Lua>,
    function: ManuallyDrop<Function<'static>>,
}

impl LuaFunction {
    pub fn new(path: &AssetPath) -> Result<Self, anyhow::Error> {
        // into_static leaks the object
        let lua = Lua::new().into_static();

        let function = ManuallyDrop::new(
            lua.load(fs::read_to_string(&path.full_path())?.as_bytes())
                .eval()?,
        );

        Ok(Self {
            lua: Some(lua),
            function,
        })
    }

    pub fn globals<F, E>(&self, mut change_globals: F) -> Result<(), anyhow::Error>
    where
        F: FnMut(&Globals<'_>) -> Result<(), E>,
        E: std::error::Error + Send + Sync + 'static,
    {
        let lua = self.lua.as_ref().unwrap();

        let globals = Globals {
            table: lua.globals(),
            lua,
        };

        change_globals(&globals)?;

        Ok(())
    }

    pub fn execute<'a, A: IntoLuaMulti<'a>, R: FromLuaMulti<'a>>(
        &self,
        input: A,
    ) -> Result<R, anyhow::Error> {
        Ok(self.function.call(input)?)
    }
}

unsafe impl Sync for LuaFunction {}
unsafe impl Send for LuaFunction {}

impl Drop for LuaFunction {
    fn drop(&mut self) {
        // enforce function before dropping lua runtime
        // to manully ensure lifetime requirements
        unsafe {
            ManuallyDrop::drop(&mut self.function);
        }

        if let Some(lua) = self.lua.take() {
            // we need to drop it, due to leakage from static cast
            let l = unsafe { Lua::from_static(lua) };
            drop(l);
        }
    }
}
